#include <u.h>
#include <libc.h>
#include <ctype.h>
#include <libg.h>
#include <lock.h>

#include "io.h"
#include "chararray.h"

#define BUFFER 256
static char output_buf[BUFFER]; /*beware shared memory!!*/
static int output_closed=0;
static int current_byte=0;
static int max_byte=0;
static int output_fd;
enum output_mode {
	vt52,
	ansi,
};
output_mode=ansi;
int wrap_mode=1;
int wrapping=0; /*1 if the last character printed was in the last column */
int saved_wrapping=0;
int scroll_area=0; /*1 if scrolling part of the screen, not all*/
int scroll_top,scroll_bottom; /* defines the scroll region */


void
output_getchar(char *utf)
{
	int i;
	Rune dummy;
	
	if (current_byte==max_byte) {
		screen_cursor(curx,cury);
		lock(&screen_lock);
		bflush();
		unlock(&screen_lock);
		max_byte=read(output_fd,output_buf,BUFFER-1);
		if (!max_byte) output_closed=1;
		current_byte=0;
		screen_cursor(curx,cury);
	}
	i=chartorune(&dummy,&output_buf[current_byte]);
	strncpy(utf,&output_buf[current_byte],i);
	if (i<UTFmax) utf[i]=0;
	current_byte+=i;
}

void
output_beep(void)
{
	int noise;

	noise=open("/dev/noise",OWRITE);
	write(noise,"800 50",6);
	close(noise);
}

void
output_check_cur(void)
{
	int x,y;

	if (curx>=maxx) {
		if (wrap_mode) {
			curx=0;  cury++;
		} else curx=maxx-1;
	}
	if (curx<0) {
		if (wrap_mode) {
			curx=maxx-1; cury--;
		} else curx=0;
	}
	if (scroll_area) {
		if (cury>scroll_bottom) {
			screen_scroll(cury-scroll_bottom,scroll_top,scroll_bottom);
			for (y=scroll_top;y<scroll_bottom-(cury-scroll_bottom);y++) for (x=0;x<maxx;x++) chararray[x][y+scroll_offset]=chararray[x][y+(cury-scroll_bottom)+scroll_offset];
			for(y=scroll_bottom;y>scroll_bottom-(cury-scroll_bottom);y--) for (x=0;x<maxx;x++) chararray[x][y+scroll_offset].character[0]='\0';
			cury=scroll_bottom;
		}
		if (cury<scroll_top) {
			screen_scroll(cury-scroll_top,scroll_top,scroll_bottom);
			for (y=scroll_bottom;y>=scroll_top+(scroll_top-cury);y--) for (x=0;x<maxx;x++) chararray[x][y+scroll_offset]=chararray[x][y-(scroll_top-cury)+scroll_offset];
			for (y=scroll_top;y<scroll_top+(scroll_top-cury);y++)  for (x=0;x<maxx;x++) chararray[x][y+scroll_offset].character[0]='\0';
			cury=scroll_top;
		}
	} else {
		if (cury>=maxy) {
			screen_scroll(cury-maxy+1,0,maxy-1);
			cury=maxy-1;
			scroll_offset+=cury-maxy+1; 
			if (scroll_offset+maxy+32>MAX_HEIGHT) {
				int x,y;

				for (y=scroll_offset-128;y<scroll_offset+maxy+2;y++) for (x=0;x<maxx;x++) chararray[x][y-scroll_offset-128]=chararray[x][y];
				scroll_offset=128;
			}
		}
		if (cury<0) { 
			screen_scroll(cury,0,maxy-1);
			/* BUG! this assumes a scroll of only one line */
			for (y=-cury;y<maxy;y++) for (x=0;x<maxx;x++) chararray[x][y+scroll_offset+1]=chararray[x][y+scroll_offset];
			for (x=0;x<maxx;x++) chararray[x][scroll_offset].character[0]='\0';
			cury=0;
		}
	}	
}


void
output_vt52(void) /*given that the ESC has been parsed, do vt52 stuff*/
{
	char c[UTFmax];

	output_getchar(&c[0]);
	switch (c[0]) {
		case '<':
			output_mode=ansi;
			break;
		default:
			/*leave the rest for now*/
			break;
	}
	return;
}

int
output_read_no(char *c)
{
	unsigned char no=0;

	do {
		no=no*10+c[0]-'0';
		output_getchar(&c[0]);
	} while (isdigit(c[0]));
	return no;
}

void
output_ansi(void) /*given that the ESC has been parsed, do ansi stuff*/
{
	char c[UTFmax],d[UTFmax];
	int x,y;

	output_getchar(&c[0]);
	switch (c[0]) {
		case '7': /*save cursor pos and attrib*/
			saved_curx=curx;
			saved_cury=cury;
			saved_wrapping=wrapping;
			saved_attrib=attrib;
			break;
		case '8': /*restore saved cursor pos and attrib*/
			curx=saved_curx;
			cury=saved_cury;
			wrapping=saved_wrapping;
			attrib=saved_attrib;
			break;
		case 'D': /*down a line, scroll at bottom*/
			cury++;
			output_check_cur();	
			break;
		case 'E': /*next line (ie LF CR), scroll at bottom*/
			curx=0; cury++; 
			output_check_cur();	
			break;
		case 'H': /*set tab stop*/
			if (wrap_mode && wrapping) tabarray[0]=1; else tabarray[curx]=1;
			break;
		case 'M': /*prev line, scroll at top*/
			cury--;
			output_check_cur();	
			break;
		case '[':
			{
				char c[UTFmax];
				int stack[16];
				int s=0,i;

				do {
					output_getchar(&c[0]);
					if (isdigit(c[0])) {
						stack[s]=output_read_no(&c[0]);
						s++;
						if (s>15) s=0; /*crude but stops crashes*/
					}
				} while (c[0]==';');
				switch (c[0]) {
					case 'A': /*cursor up*/
						cury-=s?stack[0]:1;
						if (cury<0) cury=0;
						wrapping=0;
						break;
					case 'B': /*cursor down*/
						cury+=s?stack[0]:1;
						if (cury>=maxy) cury=maxy-1;
						wrapping=0;
						break;
					case 'C': /*cursor right*/
						curx+=s?stack[0]:1;
						if (curx>=maxx) cury=maxx-1;
						wrapping=0;
						break;
					case 'D': /*cursor left*/
						curx-=s?stack[0]:1;
						if (curx<0) curx=0;
						wrapping=0;
						break;
					case 'H': /*move cursor*/
					case 'f':
						if (s>1) {
							curx=stack[1]-1;
							if (curx>=maxx) curx=maxx-1;
							if (curx<0) curx=0;
						} else curx=0;
						if (s>0) {
							cury=stack[0]-1;
							if (cury>=maxy) curx=maxy-1;
							if (cury<0) cury=0;
						} else cury=0;
						wrapping=0;
						break;
					case 'J': /*screen delete*/
						if (!s) stack[0]=0;
						switch (stack[0]) {
							case 0: /*from cursor to end of screen*/
								for (x=curx;x<maxx;x++) chararray[x][cury+scroll_offset].character[0]='\0';
								screen_delete(curx,cury,maxx-1,cury);
								for (y=cury+1;y<maxy;y++) for (x=0;x<maxx;x++) chararray[x][y+scroll_offset].character[0]='\0';
								screen_delete(0,cury+1,maxx-1,maxy-1);
								break;
							case 1: /*from cursor to start of screen*/
								for (x=0;x<=curx;x++) chararray[x][cury+scroll_offset].character[0]='\0';
								screen_delete(0,cury,curx,cury);
								for (y=0;y<cury;y++) for (x=0;x<maxx;x++) chararray[x][y+scroll_offset].character[0]='\0';
								screen_delete(0,0,maxx-1,cury-1);
								break;
							case 2: /*entire screen*/
								for (y=0;y<maxx;y++)
									for (x=0;x<maxx;x++) chararray[x][y+scroll_offset].character[0]='\0';
								screen_blank();
								break;
							default: /*invalid*/
								break;
						}
						break;
					case 'K': /*line delete*/
						if (!s) stack[0]=0;
						switch (stack[0]) {
							case 0: /*to end of line*/
								for (x=cury;x<maxx;x++) chararray[x][cury+scroll_offset].character[0]='\0';
								screen_delete(curx,cury,maxx-1,cury);
								break;
							case 1: /*from start of line*/
								for (x=0;x<=cury;x++) chararray[x][cury+scroll_offset].character[0]='\0';
								screen_delete(0,cury,curx,cury);
								break;
							case 2: /*whole line*/
								for (x=0;x<maxx;x++) chararray[x][cury+scroll_offset].character[0]='\0';
								screen_delete(0,cury,maxx-1,cury);
								break;
						}
						break;
					case 'g': /*clear tab stops*/
						if (!s) stack[0]=0;
						switch (stack[0]) {
							case 0: /*clear tab at current column*/
								if (wrap_mode && wrapping) tabarray[maxx-1]=0; else tabarray[curx]=0;
								break;
						}
						break;
					case 'L': /* insert lines */
						if (!s) {
							stack[0]=1;
						}
						for (y=maxy-1;y>=cury+stack[0];y--) for (x=0;x<maxx;x++) chararray[x][y+scroll_offset]=chararray[x][y-stack[0]+scroll_offset];
						for (y=cury;y<cury+stack[0];y++) for (x=0;x<maxx;x++) chararray[x][y+scroll_offset].character[0]='\0';
						screen_redraw(); /* do this better..*/
						break;
					case 'M': /* remove lines */
						if (!s) {
							stack[0]=1;
						}
						for (y=cury;y<=cury+stack[0];y++) for (x=0;x<maxx;x++) chararray[x][y+scroll_offset]=chararray[x][y+stack[0]+scroll_offset];
						for (y=maxy-stack[0];y<maxy;y++) for (x=0;x<maxx;x++) chararray[x][y+scroll_offset].character[0]='\0';
						screen_redraw(); /* do this better..*/
						break;
					case 'm': /*character attribs*/
						if (!s) {
							stack[0]=0;
							s=1;
						}
						for (i=0;i<s;i++) {
							switch (stack[0]) {
								case 0: /*default*/
									attrib=default_attrib;
									break;
								case 1: /* bold */
									attrib.bold=1;
									break;
								case 3: /* italics */
									attrib.italics=1;
									break;
								case 4: /* underscore */
									attrib.under=1;
									break;
								case 5: /* blink */
									attrib.blink=1;
									break;
								case 7: /* reverse video */
									attrib.reverse=1;
									break;
							}
						}
						break;
					case 'r': /* set scroll area */
						if (s>1) {
							scroll_bottom=stack[1]-1;
							if (scroll_bottom>=maxy) scroll_bottom=maxy-1;
							if (scroll_bottom<0) scroll_bottom=0;
						} else scroll_bottom=maxy-1;
						if (s>0) {
							scroll_top=stack[0]-1;
							if (scroll_top>=maxy) scroll_top=maxy-1;
							if (scroll_top<0) scroll_top=0;
						} else scroll_top=0;
						if (scroll_top>scroll_bottom) scroll_top=scroll_bottom;
						scroll_area=(scroll_top==0 && scroll_bottom==maxy-1)?0:1;
						break;
					case '?': /*mode*/
						output_getchar(&c[0]);
						output_getchar(&d[0]);
						switch (c[0]) {
							case '7': /*wrap*/
								wrap_mode=d[0]=='h'?1:0;
								break;
						}
						break;
					default: /*invalid*/
						break;
				}
			}
			break;
		default:
			break;
	}
	return;
}


void
output(int fd)
{
	char c[UTFmax];

	output_fd=fd;
	/*output_beep();*/
	sprint(status,"VTE (%dx%d)",maxx,maxy);
	screen_status(status);
	while (!output_closed) {
		output_getchar(&c[0]);
		switch (c[0]) {
			case 0: /*padding*/
				break;
			case 27:
				switch (output_mode) {
					case vt52:
						output_vt52();
						break;
					case ansi:
						output_ansi();
						break;
				}
				break;
			case 10: /*line feed*/
				wrapping=0; /*don't leave a blank line if the last line was full*/
				cury++;
				output_check_cur();	
				break;
			case 13: /* CR */
				wrapping=0;
				curx=0;
				break;
			case 8: /*backspace*/
			case 128: /*delete*/
				if (wrapping) wrapping=0; else curx--;
				output_check_cur();	
				break;
			case 9: /*tab*/
				for (curx++;curx<maxx && !tabarray[curx];curx++);
				output_check_cur();
				break;
			case 7: /*bel*/
				output_beep();
				break;
			default:
				if (c[0]>31) {	
					if (wrapping) {
						curx=0; cury++;
						wrapping=0;
						output_check_cur();	
					}
					strncpy(chararray[curx][cury+scroll_offset].character,c,UTFmax);
					chararray[curx][cury+scroll_offset].state=attrib;
					screen_plot(curx,cury,c);
					curx++;
					if (curx==maxx && wrap_mode) { wrapping=1; curx--; }
					output_check_cur();	
				}
				break;
		}
		
	}
}
