#include <u.h>
#include <libc.h>
#include <libg.h>
#include <lock.h>

#include "chararray.h"

Point fsize; /* size of chosen font */
int maxx,maxy,scroll_offset; /* size of window in characters and the position in the scrollback history */
int curx=0,cury=0; /* position of cursor in characters */
int saved_curx=0,saved_cury=0;
Char chararray[MAX_WIDTH][MAX_HEIGHT]; /* scrollback history - memory hungry, but hopefully fast */
char tabarray[MAX_WIDTH];
Bitmap *cur; /* a simple black cursor */
Bitmap *uline; /* underline */
int *x_lookup, *y_lookup; /* lookup tables for character → screen coord transation (+1 the number of characters on the screen)*/
char *status;
CharAttrib default_attrib; /*default character attributes*/
CharAttrib attrib; /*current character attributes*/
CharAttrib saved_attrib;
Font *bold, *normal;
Lock screen_lock; /*mutex on /dev/bitblt */
Rectangle sr; /*screen reactangle*/
int screen_pid; /*id of process controlling /dev/bitblt */
int screen_border;

void screen_status_u(char *s);
void screen_plot_u(int x,int y,char *s);
void screen_cursor_u(int x,int y);

void
get_size(void)
{
	maxx=(sr.max.x-sr.min.x)/fsize.x;
	if (maxx>MAX_WIDTH) maxx=MAX_WIDTH;
	maxy=(sr.max.y-sr.min.y)/fsize.y-2;
}

void
screen_blank_u(void)
{
	bitblt(&screen, screen.clipr.min, &screen, sr, Zero);
	segment(&screen,Pt(sr.min.x,sr.min.y+(fsize.y*3)/2),Pt(sr.min.x+maxx*fsize.x,sr.min.y+(fsize.y*3)/2),13,S);
	screen_status_u(status);
}

void
screen_blank(void)
{
	lock(&screen_lock);
	screen_blank_u();
	unlock(&screen_lock);
}

void
screen_redraw_u(void)
{
	int x,y;

	screen_blank_u();
	for (y=0;y<=maxy;y++) for (x=0;x<=maxx;x++) {
		Font *f=normal;
		Fcode code=notS;
		char s[UTFmax+1];

		if (chararray[x][y+scroll_offset].character[0]) {
			strncpy(s,chararray[x][y+scroll_offset].character,UTFmax);
			s[UTFmax]='\0';
			if (chararray[x][y+scroll_offset].state.bold) f=bold;
			if (chararray[x][y+scroll_offset].state.reverse) code=S;
			bitblt(&screen,Pt(x_lookup[x],y_lookup[y]),cur,cur->r,code);
			string(&screen,Pt(x_lookup[x],y_lookup[y]),f,s,DxorS);
			if (chararray[x][y+scroll_offset].state.italics || chararray[x][y+scroll_offset].state.under) {
				bitblt(&screen,Pt(x_lookup[x],y_lookup[y]),uline,uline->r,DxorS);
			}
		}
	}
	attrib=chararray[curx][cury+scroll_offset].state;
}

void
screen_redraw(void)
{
	lock(&screen_lock);
	screen_redraw_u();
	unlock(&screen_lock);
}

void
ereshaped (Rectangle r)
{
	int i;

	unlock(&screen_lock); /* insecure - but what can you do? */
	lock(&screen_lock);
	sr=inset(r,screen_border);
	get_size();
	free(x_lookup); free(y_lookup);
	x_lookup=calloc(maxx+1,sizeof(int));
	y_lookup=calloc(maxy+1,sizeof(int));
	for (i=0;i<=maxx;i++) x_lookup[i]=sr.min.x+i*fsize.x;
	for (i=0;i<=maxy;i++) y_lookup[i]=sr.min.y+(i+2)*fsize.y;
	sprint(status,"VTE (%dx%d)",maxx,maxy);
	screen_redraw_u();
	screen_cursor_u(curx,cury);
	bflush();
	unlock(&screen_lock);
	return;
}

void
screen_init(char *label)
{
	int i,k;

	lock(&screen_lock);
	binit(0,0,label);
	normal=rdfontfile("/usr/james/lib/font/latin1.8x13.font",1);
	bold=rdfontfile("/usr/james/lib/font/latin1.8x13B.font",1);
	sr=screen.clipr;
	screen_border=screen.clipr.min.x-screen.r.min.x;
	einit(Emouse|Ekeyboard);
	status=calloc(maxx,sizeof(char));
	screen_blank_u();
	fsize=strsize(normal,"M"); fsize.y=fsize.y*4/3;
	get_size();
	sprint(status,"VTE - Initialising terminal. WAIT");
	screen_blank_u(); bflush();
	for (i=0;i<MAX_HEIGHT;i++) for (k=0;k<MAX_WIDTH;k++) chararray[k][i].character[0]='\0';
	for (k=0;k<MAX_WIDTH;k++) tabarray[k]=k%8?0:1;
	default_attrib.bold=0; default_attrib.under=0; default_attrib.blink=0; default_attrib.reverse=0;
	attrib=default_attrib;
	cur=balloc(Rect(0,0,fsize.x,fsize.y),1);
	uline=balloc(Rect(0,0,fsize.x,fsize.y),1);
	x_lookup=calloc(maxx+1,sizeof(int));
	y_lookup=calloc(maxy+1,sizeof(int));
	for (i=0;i<=maxx;i++) x_lookup[i]=sr.min.x+i*fsize.x;
	for (i=0;i<=maxy;i++) y_lookup[i]=sr.min.y+(i+2)*fsize.y;
	if (!cur) {
		exits("out of bm memory");
	}
	bitblt(cur,Pt(0,0),cur,cur->r,notS);
	segment(uline,Pt(0,normal->ascent+4),Pt(fsize.x,normal->ascent+4),13,S);
	unlock(&screen_lock);
}


void
screen_plot_u(int x, int y, char *c)
{
	char s[UTFmax+1];
	Font *f=normal;
	Fcode code=notS;

	strncpy(s,c,UTFmax);
	s[UTFmax]='\0';
	if (attrib.bold) f=bold;
	if (attrib.reverse) code=S;
	bitblt(&screen,Pt(x_lookup[x],y_lookup[y]),cur,cur->r,code);
	string(&screen,Pt(x_lookup[x],y_lookup[y]),f,s,DxorS);
	if (attrib.italics || attrib.under) {
		bitblt(&screen,Pt(x_lookup[x],y_lookup[y]),uline,uline->r,DxorS);
	}
}

void
screen_plot(int x, int y, char *c)
{
	lock(&screen_lock);
	screen_plot_u(x, y, c);
	unlock(&screen_lock);
}

void
screen_status_u(char *s)
{
	bitblt(&screen, sr.min, &screen, Rect(sr.min.x,sr.min.y,sr.max.x,sr.min.y+(fsize.y*3)/2), 0);
	string(&screen,add(sr.min,Pt(0,fsize.y/2)),normal,s,S);
}

void
screen_status(char *s)
{
	lock(&screen_lock);
	screen_status_u(s);
	unlock(&screen_lock);
}

void
screen_cursor_u(int x,int y)
{
	bitblt(&screen,Pt(x_lookup[x],y_lookup[y]),cur,cur->r,DxorS);
}

void
screen_cursor(int x,int y)
{
	lock(&screen_lock);
	screen_cursor_u(x,y);
	unlock(&screen_lock);
}

void
screen_scroll_u(int lines,int top,int bottom)
{
	if (lines>0) {
		bitblt(&screen,Pt(x_lookup[0],y_lookup[top]),&screen,Rect(x_lookup[0],y_lookup[top+lines],x_lookup[maxx],y_lookup[bottom+1]),S);  /*move lines up*/
		bitblt(&screen,Pt(x_lookup[0],y_lookup[bottom+1-lines]),&screen,Rect(x_lookup[0],y_lookup[bottom+1-lines],x_lookup[maxx],y_lookup[bottom+1]),Zero); /*blank freed space*/
	} else {
		bitblt(&screen,Pt(x_lookup[0],y_lookup[top-lines]),&screen,Rect(x_lookup[0],y_lookup[top],x_lookup[maxx],y_lookup[bottom+1+lines]),S);  /*move lines down*/
		bitblt(&screen,Pt(x_lookup[0],y_lookup[top]),&screen,Rect(x_lookup[0],y_lookup[top],x_lookup[maxx],y_lookup[top-lines]),Zero); /*blank freed space*/
	}
}
#ifdef THIS_IS_NOT_DEFINED
void
screen_scroll_u(int lines)
{
	if (lines>0) {
		bitblt(&screen,Pt(x_lookup[0],y_lookup[0]),&screen,Rect(x_lookup[0],y_lookup[lines],x_lookup[maxx],y_lookup[maxy]),S);  /*move lines up*/
		bitblt(&screen,Pt(x_lookup[0],y_lookup[maxy-lines]),&screen,Rect(x_lookup[0],y_lookup[maxy-lines],x_lookup[maxx],y_lookup[maxy]),Zero); /*blank freed space*/
	} else {
		bitblt(&screen,Pt(x_lookup[0],y_lookup[0-lines]),&screen,Rect(x_lookup[0],y_lookup[0],x_lookup[maxx],y_lookup[maxy+lines]),S);  /*move lines down*/
		bitblt(&screen,Pt(x_lookup[0],y_lookup[0]),&screen,Rect(x_lookup[0],y_lookup[0],x_lookup[maxx],y_lookup[0-lines]),Zero); /*blank freed space*/
	}
}
#endif

void
screen_scroll(int lines,int top,int bottom)
{
	lock(&screen_lock);
	screen_scroll_u(lines,top,bottom);
	unlock(&screen_lock);
}

void
screen_delete_u(int x0, int y0, int x1, int y1)
{
	bitblt(&screen,Pt(x_lookup[x0],y_lookup[y0]),&screen,Rect(x_lookup[x0],y_lookup[y0],x_lookup[x1+1],y_lookup[y1+1]),Zero);
}

void
screen_delete(int x0, int y0, int x1, int y1)
{
	lock(&screen_lock);
	screen_delete_u(x0,y0,x1,y1);
	unlock(&screen_lock);
}
