#include <u.h>
#include <libc.h>
#include <libg.h>
#include <lock.h>

#include "chararray.h"
#include "io.h"

#define VERSION "1.02"
#define USAGE "Usage: vte [-?] <hostname> [ <port> | <service> ]\n"

int child_pid[10];
int chldrn=0;
int group_pid[10];
int group=0;

void
die(void)
{
	int i;
	
	for (i=0;i<chldrn;i++) {
		postnote(PNPROC,child_pid[i],"die");
	}
	for (i=0;i<group;i++) {
		postnote(PNGROUP,group_pid[i],"die");
	}	
}

int
main(int argc, char *argv[])
{
	int in_pipe[2];
	int out_pipe[2];
	int tcs_out[2];
	int f;
	char *host=0;
	char *port=0;
	char telnet_arg[128];

	atexit(die);
	ARGBEGIN {
		case '?':
			print("Viable Terminal Emulator %s\nCopyright ©1996 James Carter (james@minster.york.ac.uk)\n",VERSION);
			exits(0);
		default:
			print(USAGE);
			exits("args");
			break;
	} ARGEND
	for (;*argv;argv++) {
		if (!host) host=*argv; else {
			if (!port) port=*argv; else {
				print(USAGE);
				exits("args");
			}
		}
	}
	if (!host) {
		print(USAGE);
		exits("args");
	}
	lockinit();
	screen_init(host);
	sprint(telnet_arg,"tcp!%s!%s",host,port?port:"telnet");
	f=create("/env/TERM",OWRITE,0777);
	if (write(f,"vt100",5)!=5) {
		perror(argv0);
		exits("write");
	}
	close(f);
	if (!(pipe(in_pipe)!=-1 && pipe(out_pipe)!=-1)) {
		perror(argv0);
		exits("pipe");
	}
	if (pipe(tcs_out)==-1) {
		perror(argv0);
		exits("pipe");
	}
	switch (child_pid[chldrn++]=rfork(RFPROC|RFFDG|RFMEM)) { /* must get rid of this soon - tcs is tooooo big */
		case -1:
			perror(argv0);
			exits("fork");
		case 0:
			close(0); close(1);
			if (!(dup(out_pipe[0],0)!=-1 && dup(tcs_out[1],1)!=-1)) {
				perror(argv0);
				exits("dup");
			}
			execl("/bin/tcs","tcs","-f","latin1",0);
			break; /*tuto morté, accidento bizzaro*/
	}
	switch (child_pid[chldrn++]=rfork(RFPROC|RFFDG|RFMEM)) { 
		case -1:
			perror(argv0);
			exits("fork");
			break;
		case 0:
			output(tcs_out[0]);
			exits(0);
			break;
		default:
			break;
	}
	switch (group_pid[group++]=rfork(RFPROC|RFFDG|RFMEM|RFNOTEG)) { /*new process - all resources shared apart from file descriptors - the child is going to do fatal things to stdin etc. it's put in a different process group so we can slaughter it's children, should we wish to...*/
		case -1:
			perror(argv0);
			exits("fork");
			break;
		case 0:
			close(0); close(1); close(2);
			if (!(dup(in_pipe[1],0)!=-1 && dup(out_pipe[1],1)!=-1 && dup(out_pipe[1],2)!=-1)) {
				perror(argv0);
				exits("dup");
			}
			execl("/bin/telnet","telnet",telnet_arg,0);
			break; /*you er don't, like, get here.. right?*/
		default:
			break;
	}
	switch (child_pid[chldrn++]=rfork(RFPROC|RFFDG|RFMEM)) { 
		case -1:
			perror(argv0);
			exits("fork");
			break;
		case 0:
			input(in_pipe[0]);
			exits(0);
			break;
	}
	switch (child_pid[chldrn++]=rfork(RFPROC|RFFDG|RFMEM)) { 
		case -1:
			perror(argv0);
			exits("fork");
			break;
		case 0:
			input_mouse();
			exits(0);
			break;
	}
	wait(0);
	return 0;
}
